/*
 * lxtrap-f.a.c
 * Copyright (C) 2019, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#include <af.h>
#include <e4.h>
#include <e4fine.h>
#include <lxcall.h>
#include <lxcast.h>
#include <lxtrap-close.h>
#include <lxtrap-drive.h>
#include <lxtrap-inter.h>
#include <lxtrap-types.h>
#include <tcvset.h>

extern const void *const x1f4_a1_walk_link;

static int join_call(void **, void **, unsigned, unsigned,
		     struct dxnews_type *, struct x1f4_function_type *);
static int join_fall(void **, int *, void **, const int *, unsigned, unsigned,
		     unsigned, struct dxnews_type *);
static int line_call(void *, void *, void **);
static int miss_call(void *, int, void *, struct x1f4_dispatch_type **);
static int pick_call(void *, void *, void **);
static int pick_fall(void *, void *, void **);
static int slip_call(void *, void *, void **);
static int slip_fall(void *, void *, void **);

static int
join_call(void **track, void **input, unsigned fold, unsigned late,
	  struct dxnews_type *line_data,
	  struct x1f4_function_type *function_data)
{
    unsigned i, j;
    unsigned char *fuse;
    void **pitch, **used;

    *track++ = used = input[-1];

    used[X1f4_E4_RENDITION] = function_data;

    pitch = line_data->track;

    fuse = line_data->fuse;

    for (i = fold; i; i--) {
	j = *fuse++;
	late -= j;
	for (; j; j--) {
	    *track++ = *input++;
	}
	{
	    *track++ = *pitch++;
	}
    }
    {
	for (j = late; j; j--) {
	    *track++ = *input++;
	}
    }

    return 0;
}


static int
join_fall(void **track, int *args, void **input, const int *fill,
	  unsigned fold, unsigned late, unsigned post,
	  struct dxnews_type *line_data)
{
    const int *wind;
    int *jack;
    unsigned i, j;
    unsigned char *fuse;
    void **pitch;

    *track++ = input[-1];

    pitch = line_data->track;

    fuse = line_data->fuse;

    jack = (void *) (line_data + 1);

    wind = fill + late;

    if (post) {
	post = fold + late;
    }

    for (i = fold; i; i--) {
	j = *fuse++;
	late -= j;
	for (; j; j--) {
	    *track++ = *input++;
	    if (post) {
		args[post] = *wind++;
	    }
	    *args++ = *fill++;
	}
	if (post) {
	    args[post] = 0;
	}
	{
	    *track++ = *pitch++;
	    *args++ = *jack++;
	}
    }
    if (post) {
	for (j = late; j; j--) {
	    *track++ = *input++;
	    args[post] = *wind++;
	    *args++ = *fill++;
	}
    } else {
	for (j = late; j; j--) {
	    *track++ = *input++;
	    *args++ = *fill++;
	}
    }

    return 0;
}


static int
line_call(void *context, void *output, void **input)
{
    int status;
    unsigned bits;

    bits = ((struct dxnews_type *) context)->linetext_data->function.flags;
    if (bits & (X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST)) {
	if (output == x1f4_a1_walk_link) {
	    status = slip_fall(context, output, input);
	} else {
	    status = pick_fall(context, output, input);
	}
    } else {
	if (output == x1f4_a1_walk_link) {
	    status = slip_call(context, output, input);
	} else {
	    status = pick_call(context, output, input);
	}
    }

    return status;
}


static int
miss_call(void *text, int success, void *output,
          struct x1f4_dispatch_type **dispatch)
{
    *dispatch = NULL;

    return success;
}


static int
pick_call(void *context, void *output, void **input)
{
    int status;
    struct dxnews_type *line_data;
    struct x1f4_linetext_type *linetext_data;
    unsigned fold, late;
    void **track;

    line_data = context;

    linetext_data = line_data->linetext_data;

    late = ((struct x1f4_linetext_type *) line_data - 1)->function.count;

    fold = line_data->fold;

    track = malloc((1 + fold + late) * sizeof(void *));
    if (track) {
	join_call
	    (track, input, fold, late, line_data, &linetext_data->function);

	status = linetext_data->function.function
	    ((void *) linetext_data->context, output, track + 1);

	free(track);
    } else {
	status = _libx1f4i0_lxtrap_stat_link(line_data->lxtrap_data);
    }

    return status;
}


static int
pick_fall(void *context, void *output, void **input)
{
    int status;
    struct dxnews_type *line_data;
    struct x1f4_function_type *function_data;
    struct x1f4_linetext_type *linetext_data;
    unsigned fold, late, post, size;
    void **used, **track;

    used = input[-1];

    function_data = used[X1f4_E4_RENDITION];

    line_data = context;

    linetext_data = line_data->linetext_data;

    post = linetext_data->function.flags
	& (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE) ? 1 : 0;

    late = function_data->count;

    fold = line_data->fold;

    size = fold + late;

    track = malloc((1 + size) * sizeof(void *) + (size * sizeof(int) << post));
    if (track) {
	int *args;
	struct x1f4_function_type function;

	args = (void *) (track + 1 + size);

	join_fall
	    (track, args, input, function_data->args, fold, late, post,
	     line_data);

	function = linetext_data->function;

	function.args = args;
	function.count = size;
	function.flags &= ~(X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST);

	if (post) {
	    function.flags |= X1f4_E4_POST_TYPE;
	}

	used[X1f4_E4_RENDITION] = &function;

	status = linetext_data->function.function
	    ((void *) linetext_data->context, output, track + 1);

	free(track);
    } else {
	status = _libx1f4i0_lxtrap_stat_link(line_data->lxtrap_data);
    }

    return status;
}


static int
slip_call(void *context, void *output, void **input)
{
    int status;
    struct dxnews_type *line_data;
    struct lxtrap_type *lxtrap_data;
    unsigned fold, late;
    void *patch;

    line_data = context;

    lxtrap_data = line_data->lxtrap_data;

    late = ((struct x1f4_linetext_type *) line_data - 1)->function.count;

    fold = line_data->fold;

    status = lxtrap_data->link_v.link
	(lxtrap_data->link_v.data, &patch,
	 sizeof(struct x1f4_dispatch_type)
	 + (fold + late + 1) * sizeof(void *));
    if (status) {
    } else {
	struct x1f4_dispatch_type *dispatch_data;
	struct x1f4_linetext_type *linetext_data;
	void **track;

	linetext_data = line_data->linetext_data;

	dispatch_data = patch;

	track = (void *) (dispatch_data + 1);

	join_call
	    (track, input[1], fold, late, line_data, &linetext_data->function);

	dispatch_data->back = NULL;
	dispatch_data->call = miss_call;

	dispatch_data->transfer.context = (void *) linetext_data->context;
	dispatch_data->transfer.function_data = &linetext_data->function;
	dispatch_data->transfer.input = track + 1;
	dispatch_data->transfer.output = input[0];

	input[2] = dispatch_data;
    }

    return status;
}


static int
slip_fall(void *context, void *output, void **input)
{
    int status;
    struct dxnews_type *line_data;
    struct lxtrap_type *lxtrap_data;
    struct x1f4_function_type *function_data;
    struct x1f4_linetext_type *linetext_data;
    unsigned fold, late, post, size;
    void *patch, **phase, **used;

    phase = input[1];

    used = phase[-1];

    function_data = used[X1f4_E4_RENDITION];

    line_data = context;

    lxtrap_data = line_data->lxtrap_data;

    linetext_data = line_data->linetext_data;

    post = linetext_data->function.flags
	& (X1f4_E4_LINK_PASS | X1f4_E4_POST_TYPE) ? 1 : 0;

    late = function_data->count;

    fold = line_data->fold;

    size = fold + late;

    status = lxtrap_data->link_v.link
	(lxtrap_data->link_v.data, &patch,
	 sizeof(struct x1f4_dispatch_type) + sizeof(struct x1f4_function_type)
	 + (size + 1) * sizeof(void *) + (size * sizeof(int) << post));
    if (status) {
    } else {
	int *args;
	struct x1f4_dispatch_type *dispatch_data;
	struct x1f4_function_type *function_slip;
	void **track;

	dispatch_data = patch;

	function_slip = (void *) (dispatch_data + 1);

	track = (void *) (function_slip + 1);

	args = (void *) (track + 1 + size);

	join_fall
	    (track, args, phase, function_data->args, fold, late, post,
	     line_data);

	*function_slip = linetext_data->function;

	function_slip->args = args;
	function_slip->count = size;
	function_slip->flags &= ~(X1f4_E4_SIDE_LIST | X1f4_E4_SLIP_LIST);

	if (post) {
	    function_slip->flags |= X1f4_E4_POST_TYPE;
	}

	used[X1f4_E4_RENDITION] = function_slip;

	dispatch_data->back = NULL;
	dispatch_data->call = miss_call;

	dispatch_data->transfer.context = (void *) linetext_data->context;
	dispatch_data->transfer.function_data = &linetext_data->function;
	dispatch_data->transfer.input = track + 1;
	dispatch_data->transfer.output = input[0];

	input[2] = dispatch_data;
    }

    return status;
}


int
_libx1f4i0_lxtrap_a_forward(void *context, void *output, void **input)
{
    int linked = 0, status = 0;
    struct dxnews_type *line_data = NULL;
    struct lxtrap_type *lxtrap_data;
    struct x1f4_leadnode_type *leadnode_data = NULL;
    struct x1f4_linetext_type *linetext_data;
    unsigned char *rule = NULL;

    lxtrap_data = context;

    do {
	int *args, *edge, *suit, type;
	const int *fill, *such;
	struct frame_type frame;
	struct x1f4_dxcast_type *dxcast_data;
	struct x1f4_function_type function[2], *function_data;
	struct x1f4_linetext_type *linetext_over;
	struct x1f4_nodetype_type *nodetype_data;
	unsigned fold, i, late, none = X1f4_E4_SLIP_LIST, mill, over, post,
	    rise;
	unsigned char *fuse;
	void *line, **order, *spin, **track, **used;

	dxcast_data = I_USER(input[0]);

	if (X1f4_E4_CALL < dxcast_data->type) {
	} else {
	    status = _libx1f4i0_lxtrap_stat_roll(lxtrap_data);
	    if (1) {
		break;
	    }
	}

	linetext_data = I_USER(&dxcast_data->data);
	if (linetext_data->function.function) {
	} else {
	    status = _libx1f4i0_lxtrap_stat_zero(lxtrap_data);
	    if (1) {
		break;
	    }
	}

	used = input[-1];

	function_data = used[X1f4_E4_RENDITION];

	fold = function_data->count - 1;

	if (fold) {
	} else {
	    l_USER(dxcast_data, output);

	    status = 0;

	    break;
	}

	if (fold & 1) {
	    status = _libx1f4i0_lxtrap_stat_even(lxtrap_data);
	    if (1) {
		break;
	    }
	}

	if (254 < fold) {
	    status = _libx1f4i0_lxtrap_stat_over(lxtrap_data);
	    if (1) {
		break;
	    }
	}

	over = linetext_data->function.flags;

	mill = linetext_data->function.count;

	late = mill;

	rule = malloc(mill);
	if (rule) {
	} else {
	    status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
	    if (1) {
		break;
	    }
	}

	fill = function_data->args - 1;

	order = input - 1;

	memset(rule, 0, mill);

	fold += 2;
	for (i = 2; i < fold; i += 2) {
	    X1f4_E4_C_MODE mode;

	    fill += 2;

	    type = *fill;
	    if (type != X1f4_E4_MODE) {
		status = _libx1f4i0_lxtrap_stat_iran(context, i - 1);
		if (1) {
		    break;
		}
	    }

	    order += 2;

	    mode = I_MODE(*order);
	    if (mode < 0) {
		if (mode < -(int) mill) {
		    status = _libx1f4i0_lxtrap_stat_slip(context, mill, mode);
		    if (1) {
			break;
		    }
		}

		mode += mill;
	    } else {
		if (mill <= mode) {
		    status = _libx1f4i0_lxtrap_stat_slip(context, mill, mode);
		    if (1) {
			break;
		    }
		}
	    }

	    if (rule[mode]) {
	    } else {
		late--;
	    }

	    rule[mode] = i;
	}
	if (status) {
	    break;
	}

	fold = mill - late;

	such = linetext_data->function.args;

	post = over & X1f4_E4_POST_TYPE ? 1 : 0;

	if (post) {
	    const int *pass;

	    pass = such + mill;
	    i = mill;
	    do {
		i--;
		if (rule[i] && pass[i] & X1f4_E4_POST_XSET) {
		    status = _libx1f4i0_lxcall_stat_fink
			(&lxtrap_data->link_e, i, &linetext_data->function);
		    if (1) {
			break;
		    }
		}
	    } while (i);
	    if (status) {
		break;
	    }
	}

	rise = 0;
	for (i = mill; i; i--) {
	    if (*rule++) {
		if (255 < rise) {
		    status = _libx1f4i0_lxtrap_stat_rise(context);
		    if (1) {
			break;
		    }
		} else {
		    rise = 0;
		}
	    } else {
		rise++;
	    }
	}

	rule += i;
	rule -= mill;

	if (status) {
	    break;
	}

	status = lxtrap_data->link_m.link
	    (lxtrap_data->link_m.data, &line,
	     sizeof(struct x1f4_leadnode_type)
	     + sizeof(struct x1f4_linetext_type)
	     + sizeof(dxnews_type) + (sizeof(int) * (fold + (late << post)))
	     + fold);
	if (status) {
	    status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
	    if (1) {
		break;
	    }
	}

	leadnode_data = line;

	linetext_over = (void *) (leadnode_data + 1);

	line = linetext_over + 1;

	line_data = line;

	leadnode_data->line.call = 0;

	_libx1f4i0_init_tcvset
	    (&leadnode_data->line.pset, &leadnode_data->line.text);

	line_data->lxtrap_data = lxtrap_data;

	*linetext_over = *linetext_data;

	args = (void *) (line_data + 1);

	suit = args + fold;

	fuse = (void *) (suit + (late << post));

	line_data->fuse = fuse;

	line_data->track = NULL;

	track = spin = malloc((sizeof(void *) + sizeof(int)) * fold);
	if (track) {
	} else {
	    status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
	    if (1) {
		break;
	    }
	}

	fill = function_data->args;

	edge = (void *) (track + fold);

	rise = 0;
	for (i = mill; i; i--) {
	    unsigned j;

	    type = *such;
	    j = *rule++;
	    if (j) {
		if (type == X1f4_E4_SLIP) {
		    type = function_data->args[j];
		}

		*args++ = type;
		*edge++ = fill[j];
		*track++ = input[j];
		*fuse++ = rise;
		rise = 0;
	    } else {
		*suit = type;
		if (post) {
		    suit[late] = such[mill];
		}
		suit++;
		rise++;
		if (type == X1f4_E4_SLIP) {
		    none = 0;
		}
	    }

	    such++;
	}

	free(rule - mill);

	rule = NULL;

	args -= fold;

	track -= fold;

	linetext_over->context = line_data;

	linetext_over->function.args = args + fold;
	linetext_over->function.count = late;
	linetext_over->function.function = line_call;
	linetext_over->function.length = 7;
	linetext_over->function.name = "applied";
	linetext_over->function.type = linetext_data->function.type;

	linetext_over->function.flags = X1f4_E4_OVER_LINK | X1f4_E4_TEXT_LINK
	    | (over & ~(X1f4_E4_LEAD_CALL | X1f4_E4_NEWS_LINK
			| X1f4_E4_NODE_LINK));
	linetext_over->function.flags &= ~none;
	if (over & (X1f4_E4_LEAD_CALL | X1f4_E4_NODE_LINK)||1) {
	    linetext_over->function.flags |= X1f4_E4_WALK_LINK;
	}

	nodetype_data = lxtrap_data->link_x.fold[0].nodetype;

	leadnode_data->link.nodetype_data = nodetype_data;
	leadnode_data->link.slip = NULL;

	frame.data = lxtrap_data->link_m.data;
	frame.link = lxtrap_data->link_m.link;

	function[0] = linetext_data->function;
	function[0].args = args;
	function[0].flags &= ~X1f4_E4_POST_TYPE;
	function[0].count = fold;

	function[1] = linetext_data->function;
	function[1].args = edge - fold;
	function[1].flags &= ~X1f4_E4_POST_TYPE;
	function[1].count = fold;

	status = _libx1f4i0_lxcall_even_list
	    (function, function + 1, 0, &track,
	     X1f4_LXCALL_POST_COPY | X1f4_LXCALL_POST_HOLD
	     /* | X1f4_LXCALL_POST_WAYS */,
	     &frame, &lxtrap_data->link_p, &lxtrap_data->link_q,
	     &lxtrap_data->link_e, lxtrap_data->trap_e.eelookup_data);

	free(spin);

	spin = NULL;

	if (status) {
	    break;
	}

	line_data->track = track;

	status = _x1f4_node_effect
	    (NULL, (void *) &line_data->linetext_data, linetext_data,
	     &leadnode_data->link);
	if (status) {
	    break;
	}

	linked = 1;

	status = _libx1f4i0_lxtrap_node_args
	    (line_data, args, fold, &leadnode_data->link);
	if (status) {
	    if (1) {
		break;
	    }
	}

	status = x1f4_type_e4fine
	    (lxtrap_data->link_o.fine, &type, linetext_over);
	if (status) {
	    status = _libx1f4i0_lxtrap_stat_link(lxtrap_data);
	    if (1) {
		break;
	    }
	}

	status = x1f4_cast_node
	    (&lxtrap_data->link_v, &lxtrap_data->link_e, nodetype_data, type,
	     &linetext_over, output);
    } while (0);

    if (status) {
	if (leadnode_data) {
	    if (line_data->track) {
		if (linked) {
		    if (line_data->fold) {
			_libx1f4i0_lxtrap_free_args(line_data, NULL);
		    }

		    _x1f4_free_effect
			(NULL, linetext_data, NULL, &leadnode_data->link);
		}

		lxtrap_data->link_m.free
		    (lxtrap_data->link_m.data, line_data->track - 1);
	    }

	    lxtrap_data->link_m.free
		(lxtrap_data->link_m.data, leadnode_data);
	}
	if (rule) {
	    free(rule);
	}
    }

    return status;
}
